"""
Interface wraps quantum function as a jax function
"""

from typing import Any, Callable, Tuple, Optional, Union, Sequence
from functools import wraps, partial

from ..cons import backend
from .tensortrans import general_args_to_backend

Tensor = Any


def jax_wrapper(
    fun: Callable[..., Any],
    enable_dlpack: bool = False,
    output_shape: Optional[
        Union[Tuple[int, ...], Tuple[int, ...], Sequence[Tuple[int, ...]]]
    ] = None,
    output_dtype: Optional[Union[Any, Sequence[Any]]] = None,
) -> Callable[..., Any]:
    import jax

    @wraps(fun)
    def fun_jax(*x: Any) -> Any:
        def wrapped_fun(*args: Any) -> Any:
            args = general_args_to_backend(args, enable_dlpack=enable_dlpack)
            y = fun(*args)
            y = general_args_to_backend(
                y, target_backend="jax", enable_dlpack=enable_dlpack
            )
            return y

        # Use provided shape and dtype if available, otherwise run test
        if output_shape is not None and output_dtype is not None:
            if (
                isinstance(output_shape, Sequence)
                and len(output_shape) > 0
                and not isinstance(output_shape[0], int)
            ):
                # Multiple outputs case
                out_shape = tuple(
                    jax.ShapeDtypeStruct(s, d)
                    for s, d in zip(output_shape, output_dtype)
                )
            else:
                # Single output case
                out_shape = jax.ShapeDtypeStruct(output_shape, output_dtype)  # type: ignore
        else:
            # Get expected output shape by running function once
            test_out = wrapped_fun(*x)
            if isinstance(test_out, tuple):
                # Multiple outputs case
                out_shape = tuple(
                    jax.ShapeDtypeStruct(
                        t.shape if hasattr(t, "shape") else (),
                        t.dtype if hasattr(t, "dtype") else x[0].dtype,
                    )
                    for t in test_out
                )
            else:
                # Single output case
                out_shape = jax.ShapeDtypeStruct(  # type: ignore
                    test_out.shape if hasattr(test_out, "shape") else (),
                    test_out.dtype if hasattr(test_out, "dtype") else x[0].dtype,
                )

        # Use pure_callback with correct output shape
        result = jax.pure_callback(wrapped_fun, out_shape, *x)
        return result

    return fun_jax


def jax_interface(
    fun: Callable[..., Any],
    jit: bool = False,
    enable_dlpack: bool = False,
    output_shape: Optional[Union[Tuple[int, ...], Tuple[()]]] = None,
    output_dtype: Optional[Any] = None,
) -> Callable[..., Any]:
    """
    Wrap a function on different ML backend with a jax interface.

    :Example:

    .. code-block:: python

        tc.set_backend("tensorflow")

        def f(params):
            c = tc.Circuit(1)
            c.rx(0, theta=params[0])
            c.ry(0, theta=params[1])
            return tc.backend.real(c.expectation([tc.gates.z(), [0]]))

        f = tc.interfaces.jax_interface(f, jit=True)

        params = jnp.ones(2)
        value, grad = jax.value_and_grad(f)(params)

    :param fun: The quantum function with tensor in and tensor out
    :type fun: Callable[..., Any]
    :param jit: whether to jit ``fun``, defaults to False
    :type jit: bool, optional
    :param enable_dlpack: whether transform tensor backend via dlpack, defaults to False
    :type enable_dlpack: bool, optional
    :param output_shape: Optional shape of the function output, defaults to None
    :type output_shape: Optional[Union[Tuple[int, ...], Tuple[()]]], optional
    :param output_dtype: Optional dtype of the function output, defaults to None
    :type output_dtype: Optional[Any], optional
    :return: The same quantum function but now with jax array in and jax array out
        while AD is also supported
    :rtype: Callable[..., Any]
    """
    jax_fun = create_jax_function(
        fun,
        enable_dlpack=enable_dlpack,
        jit=jit,
        output_shape=output_shape,
        output_dtype=output_dtype,
    )
    return jax_fun


def create_jax_function(
    fun: Callable[..., Any],
    enable_dlpack: bool = False,
    jit: bool = False,
    output_shape: Optional[Union[Tuple[int, ...], Tuple[()]]] = None,
    output_dtype: Optional[Any] = None,
) -> Callable[..., Any]:
    import jax
    from jax import custom_vjp

    if jit:
        fun = backend.jit(fun)

    wrapped = jax_wrapper(
        fun,
        enable_dlpack=enable_dlpack,
        output_shape=output_shape,
        output_dtype=output_dtype,
    )

    @custom_vjp
    def f(*x: Any) -> Any:
        return wrapped(*x)

    def f_fwd(*x: Any) -> Tuple[Any, Tuple[Any, ...]]:
        y = wrapped(*x)
        return y, x

    def f_bwd(res: Tuple[Any, ...], g: Any) -> Tuple[Any, ...]:
        x = res

        if len(x) == 1:
            x = x[0]

        vjp_fun = partial(backend.vjp, fun)
        if jit:
            vjp_fun = backend.jit(vjp_fun)  # type: ignore

        def vjp_wrapped(args: Any) -> Any:
            args = general_args_to_backend(args, enable_dlpack=enable_dlpack)
            gb = general_args_to_backend(g, enable_dlpack=enable_dlpack)
            r = vjp_fun(args, gb)[1]
            r = general_args_to_backend(
                r, target_backend="jax", enable_dlpack=enable_dlpack
            )
            return r

        # Handle gradient shape for both single input and tuple inputs
        if isinstance(x, tuple):
            # Create a tuple of ShapeDtypeStruct for each input
            grad_shape = tuple(jax.ShapeDtypeStruct(xi.shape, xi.dtype) for xi in x)
        else:
            grad_shape = jax.ShapeDtypeStruct(x.shape, x.dtype)

        dx = jax.pure_callback(
            vjp_wrapped,
            grad_shape,
            x,
        )

        if not isinstance(dx, tuple):
            dx = (dx,)
        return dx  # type: ignore

    f.defvjp(f_fwd, f_bwd)
    return f



"""
Interface wraps quantum function as a torch function
"""

from typing import Any, Callable, Dict, Tuple
from functools import partial

from ..cons import backend
from ..utils import is_sequence
from .tensortrans import general_args_to_backend

Tensor = Any

# TODO(@refraction-ray): new paradigm compatible with torch functional transformation


def torch_interface(
    fun: Callable[..., Any], jit: bool = False, enable_dlpack: bool = False
) -> Callable[..., Any]:
    """
    Wrap a quantum function on different ML backend with a pytorch interface.

    :Example:

    .. code-block:: python

        import torch

        tc.set_backend("tensorflow")


        def f(params):
            c = tc.Circuit(1)
            c.rx(0, theta=params[0])
            c.ry(0, theta=params[1])
            return c.expectation([tc.gates.z(), [0]])


        f_torch = tc.interfaces.torch_interface(f, jit=True)

        a = torch.ones([2], requires_grad=True)
        b = f_torch(a)
        c = b ** 2
        c.backward()

        print(a.grad)

    :param fun: The quantum function with tensor in and tensor out
    :type fun: Callable[..., Any]
    :param jit: whether to jit ``fun``, defaults to False
    :type jit: bool, optional
    :param enable_dlpack: whether transform tensor backend via dlpack, defaults to False
    :type enable_dlpack: bool, optional
    :return: The same quantum function but now with torch tensor in and torch tensor out
        while AD is also supported
    :rtype: Callable[..., Any]
    """
    import torch

    def vjp_fun(x: Tensor, v: Tensor) -> Tuple[Tensor, Tensor]:
        return backend.vjp(fun, x, v)

    if jit is True:
        fun = backend.jit(fun)
        vjp_fun = backend.jit(vjp_fun)

    class Fun(torch.autograd.Function):  # type: ignore
        @staticmethod
        def forward(ctx: Any, *x: Any) -> Any:  # type: ignore
            # ctx.xdtype = [xi.dtype for xi in x]
            ctx.save_for_backward(*x)
            x_detached = backend.tree_map(lambda s: s.detach(), x)
            ctx.xdtype = backend.tree_map(lambda s: s.dtype, x_detached)
            # (x, )
            if len(ctx.xdtype) == 1:
                ctx.xdtype = ctx.xdtype[0]
            ctx.device = (backend.tree_flatten(x_detached)[0][0]).device
            x = general_args_to_backend(x_detached, enable_dlpack=enable_dlpack)
            y = fun(*x)
            ctx.ydtype = backend.tree_map(lambda s: s.dtype, y)
            if len(x) == 1:
                x = x[0]
            ctx.x = x
            y = general_args_to_backend(
                y, target_backend="pytorch", enable_dlpack=enable_dlpack
            )
            y = backend.tree_map(lambda s: s.to(device=ctx.device), y)
            return y

        @staticmethod
        def backward(ctx: Any, *grad_y: Any) -> Any:
            x = ctx.saved_tensors
            x_detached = backend.tree_map(lambda s: s.detach(), x)
            x_backend = general_args_to_backend(x_detached, enable_dlpack=enable_dlpack)
            if len(grad_y) == 1:
                grad_y = grad_y[0]
            grad_y = backend.tree_map(lambda s: s.contiguous(), grad_y)
            grad_y = general_args_to_backend(
                grad_y, dtype=ctx.ydtype, enable_dlpack=enable_dlpack
            )
            # grad_y = general_args_to_numpy(grad_y)
            # grad_y = numpy_args_to_backend(grad_y, dtype=ctx.ydtype)  # backend.dtype
            if len(x_backend) == 1:
                x_backend_for_vjp = x_backend[0]
            else:
                x_backend_for_vjp = x_backend

            _, g = vjp_fun(x_backend_for_vjp, grad_y)
            # a redundency due to current vjp API

            r = general_args_to_backend(
                g,
                dtype=ctx.xdtype,
                target_backend="pytorch",
                enable_dlpack=enable_dlpack,
            )
            r = backend.tree_map(lambda s: s.to(device=ctx.device), r)
            if not is_sequence(r):
                return (r,)
            return r

    # currently, memory transparent dlpack in these ML framework has broken support on complex dtypes
    return Fun.apply  # type: ignore


pytorch_interface = torch_interface


def torch_interface_kws(
    f: Callable[..., Any], jit: bool = True, enable_dlpack: bool = False
) -> Callable[..., Any]:
    """
    similar to py:meth:`tensorcircuit.interfaces.torch.torch_interface`,
    but now the interface support static arguments for function ``f``,
    which is not a tensor and can be used with keyword arguments

    :Example:

    .. code-block:: python

        tc.set_backend("tensorflow")

        def f(tensor, integer):
            r = 0.
            for i in range(integer):
                r += tensor
            return r

        fnew = tc.interfaces.torch_interface_kws(f)

        print(fnew(torch.ones([2]), integer=3))
        print(fnew(torch.ones([2]), integer=4))

    :param f: _description_
    :type f: Callable[..., Any]
    :param jit: _description_, defaults to True
    :type jit: bool, optional
    :param enable_dlpack: _description_, defaults to False
    :type enable_dlpack: bool, optional
    :return: _description_
    :rtype: Callable[..., Any]
    """
    cache_dict: Dict[Tuple[Any, ...], Callable[..., Any]] = {}

    def wrapper(*args: Any, **kws: Any) -> Any:
        key = tuple([(k, v) for k, v in kws.items()])
        if key not in cache_dict:
            fnew = torch_interface(
                partial(f, **kws), jit=jit, enable_dlpack=enable_dlpack
            )
            cache_dict[key] = fnew
        return cache_dict[key](*args)

    return wrapper



"""
Experimental features
"""

# pylint: disable=unused-import

from functools import partial
import logging
from typing import Any, Callable, Dict, Optional, Tuple, List, Sequence, Union
import pickle
import uuid
import time
import os

import numpy as np

from .cons import backend, dtypestr, rdtypestr, get_tn_info
from .gates import Gate
from .timeevol import hamiltonian_evol, evol_global, evol_local

# for backward compatibility

Tensor = Any
Circuit = Any

logger = logging.getLogger(__name__)


def adaptive_vmap(
    f: Callable[..., Any],
    vectorized_argnums: Union[int, Sequence[int]] = 0,
    static_argnums: Optional[Union[int, Sequence[int]]] = None,
    chunk_size: Optional[int] = None,
) -> Callable[..., Any]:
    """
    Vectorized map with adaptive chunking for memory efficiency.

    :param f: Function to be vectorized
    :param vectorized_argnums: Arguments to be vectorized over
    :param static_argnums: Arguments that remain static during vectorization
    :param chunk_size: Size of chunks for batch processing, None means no chunking
        (naive vmap)
    :return: Vectorized function
    """
    if chunk_size is None:
        return backend.vmap(f, vectorized_argnums)  # type: ignore

    if isinstance(vectorized_argnums, int):
        vectorized_argnums = (vectorized_argnums,)

    def wrapper(*args: Any, **kws: Any) -> Tensor:
        # only support `f` outputs a tensor
        s1, s2 = divmod(args[vectorized_argnums[0]].shape[0], chunk_size)  # type: ignore
        # repetition, rest
        reshape_args = []
        rest_args = []
        for i, arg in enumerate(args):
            if i in vectorized_argnums:  # type: ignore
                if s2 != 0:
                    arg_rest = arg[-s2:]
                    arg = arg[:-s2]
                arg = backend.reshape(
                    arg,
                    [s1, chunk_size] + list(backend.shape_tuple(arg))[1:],
                )

            else:
                arg_rest = arg
            reshape_args.append(arg)
            if s2 != 0:
                rest_args.append(arg_rest)
        _vmap = backend.jit(
            backend.vmap(f, vectorized_argnums=vectorized_argnums),
            static_argnums=static_argnums,
        )
        r = []
        for i in range(s1):
            # currently using naive python loop for simplicity
            nreshape_args = [
                a[i] if j in vectorized_argnums else a  # type: ignore
                for j, a in enumerate(reshape_args)
            ]
            r.append(_vmap(*nreshape_args, **kws))
        r = backend.tree_map(lambda *x: backend.concat(x), *r)
        # rshape = list(backend.shape_tuple(r))
        # if len(rshape) == 2:
        #     nshape = [rshape[0] * rshape[1]]
        # else:
        #     nshape = [rshape[0] * rshape[1], -1]
        # r = backend.reshape(r, nshape)
        if s2 != 0:
            rest_r = _vmap(*rest_args, **kws)
            return backend.tree_map(lambda *x: backend.concat(x), r, rest_r)
        return r

    return wrapper


def _qng_post_process(t: Tensor, eps: float = 1e-4) -> Tensor:
    t += eps * backend.eye(t.shape[0], dtype=t.dtype)
    t = backend.real(t)
    return t


def _id(a: Any) -> Any:
    return a


def _vdot(i: Tensor, j: Tensor) -> Tensor:
    return backend.tensordot(backend.conj(i), j, 1)


def qng(
    f: Callable[..., Tensor],
    kernel: str = "qng",
    postprocess: Optional[str] = "qng",
    mode: str = "fwd",
) -> Callable[..., Tensor]:
    """
    Compute quantum natural gradient for quantum circuit optimization.

    :param f: Function that takes parameters and returns quantum state
    :param kernel: Type of kernel to use ("qng" or "dynamics"), the former has the second term
    :param postprocess: Post-processing method ("qng" or None)
    :param mode: Mode of differentiation ("fwd" or "rev")
    :return: Function computing QNG matrix

    :Example:

    >>> import tensorcircuit as tc
    >>> def ansatz(params):
    ...     c = tc.Circuit(2)
    ...     c.rx(0, theta=params[0])
    ...     c.ry(1, theta=params[1])
    ...     return c.state()
    >>> qng_fn = tc.experimental.qng(ansatz)
    >>> params = tc.array_to_tensor([0.5, 0.8])
    >>> qng_matrix = qng_fn(params)
    >>> print(qng_matrix.shape)  # (2, 2)
    """

    # for both qng and qng2 calculation, we highly recommended complex-dtype but real valued inputs
    def wrapper(params: Tensor, **kws: Any) -> Tensor:
        params = backend.cast(params, dtype=dtypestr)  # R->C protection
        psi = f(params)
        if mode == "fwd":
            jac = backend.jacfwd(f)(params)
        else:  # "rev"
            jac = backend.jacrev(f)(params)
            jac = backend.cast(jac, dtypestr)  # incase input is real
            # may have R->C issue for rev mode, which we obtain a real Jacobian
        jac = backend.transpose(jac)
        if kernel == "qng":

            def ij(i: Tensor, j: Tensor) -> Tensor:
                return _vdot(i, j) - _vdot(i, psi) * _vdot(psi, j)

        elif kernel == "dynamics":

            def ij(i: Tensor, j: Tensor) -> Tensor:
                return _vdot(i, j)

        vij = backend.vmap(ij, vectorized_argnums=0)
        vvij = backend.vmap(vij, vectorized_argnums=1)

        fim = vvij(jac, jac)
        # TODO(@refraction-ray): investigate more on
        # suitable hyperparameters and methods for regularization?
        if isinstance(postprocess, str):
            if postprocess == "qng":
                _post_process = _qng_post_process
            else:
                raise ValueError("Unsupported postprocess option")

        elif postprocess is None:
            _post_process = _id  # type: ignore
        else:
            _post_process = postprocess  # callable
        fim = _post_process(fim)
        return fim

    return wrapper


dynamics_matrix = partial(qng, kernel="dynamics", postprocess=None)


def qng2(
    f: Callable[..., Tensor],
    kernel: str = "qng",
    postprocess: Optional[str] = "qng",
    mode: str = "rev",
) -> Callable[..., Tensor]:
    # reverse mode has a slightly better running time
    # wan's approach for qng
    def wrapper(params: Tensor, **kws: Any) -> Tensor:
        params2 = backend.copy(params)
        params2 = backend.stop_gradient(params2)

        def outer_loop(params2: Tensor) -> Tensor:
            def inner_product(params: Tensor, params2: Tensor) -> Tensor:
                s = f(params)
                s2 = f(params2)
                fid = _vdot(s2, s)
                if kernel == "qng":
                    fid -= _vdot(s2, backend.stop_gradient(s)) * _vdot(
                        backend.stop_gradient(s2), s
                    )
                return fid

            _, grad = backend.vjp(
                partial(inner_product, params2=params2), params, backend.ones([])
            )
            return grad

        if mode == "fwd":
            fim = backend.jacfwd(outer_loop)(params2)
        else:
            fim = backend.jacrev(outer_loop)(params2)
        # directly real if params is real, then where is the imaginary part?
        if isinstance(postprocess, str):
            if postprocess == "qng":
                _post_process = _qng_post_process
            else:
                raise ValueError("Unsupported postprocess option")

        elif postprocess is None:
            _post_process = _id  # type: ignore
        else:
            _post_process = postprocess  # callable
        fim = _post_process(fim)
        return fim

    # on jax backend, qng and qng2 output is different by a conj
    # on tf backend, the outputs are the same
    return wrapper


def dynamics_rhs(f: Callable[..., Any], h: Tensor) -> Callable[..., Any]:
    # compute :math:`\langle \psi \vert H \vert \partial \psi \rangle`
    def wrapper(params: Tensor, **kws: Any) -> Tensor:
        def energy(params: Tensor) -> Tensor:
            w = f(params, **kws)
            wr = backend.stop_gradient(w)
            wl = backend.conj(w)
            wl = backend.reshape(wl, [1, -1])
            wr = backend.reshape(wr, [-1, 1])
            if not backend.is_sparse(h):
                e = wl @ h @ wr
            else:
                tmp = backend.sparse_dense_matmul(h, wr)
                e = wl @ tmp
            return backend.real(e)[0, 0]

        return backend.grad(energy)(params)

    return wrapper


def parameter_shift_grad(
    f: Callable[..., Tensor],
    argnums: Union[int, Sequence[int]] = 0,
    jit: bool = False,
    shifts: Tuple[float, float] = (np.pi / 2, 2),
) -> Callable[..., Tensor]:
    """
    similar to `grad` function but using parameter shift internally instead of AD,
    vmap is utilized for evaluation, so the speed is still ok

    :param f: quantum function with weights in and expectation out
    :type f: Callable[..., Tensor]
    :param argnums: label which args should be differentiated,
        defaults to 0
    :type argnums: Union[int, Sequence[int]], optional
    :param jit: whether jit the original function `f` at the beginning,
        defaults to False
    :type jit: bool, optional
    :param shifts: two floats for the delta shift on the numerator and dominator,
        defaults to (pi/2, 2) for parameter shift
    :type shifts: Tuple[float, float]
    :return: the grad function
    :rtype: Callable[..., Tensor]
    """
    if jit is True:
        f = backend.jit(f)

    if isinstance(argnums, int):
        argnums = [argnums]

    vfs = [backend.vmap(f, vectorized_argnums=i) for i in argnums]

    def grad_f(*args: Any, **kws: Any) -> Any:
        grad_values = []
        for i in argnums:  # type: ignore
            shape = backend.shape_tuple(args[i])
            size = backend.sizen(args[i])
            onehot = backend.eye(size)
            onehot = backend.cast(onehot, args[i].dtype)
            onehot = backend.reshape(onehot, [size] + list(shape))
            onehot = shifts[0] * onehot
            nargs = list(args)
            arg = backend.reshape(args[i], [1] + list(shape))
            batched_arg = backend.tile(arg, [size] + [1 for _ in shape])
            nargs[i] = batched_arg + onehot
            nargs2 = list(args)
            nargs2[i] = batched_arg - onehot
            r = (vfs[i](*nargs, **kws) - vfs[i](*nargs2, **kws)) / shifts[1]
            r = backend.reshape(r, shape)
            grad_values.append(r)
        if len(argnums) > 1:  # type: ignore
            return tuple(grad_values)
        return grad_values[0]

    return grad_f


def parameter_shift_grad_v2(
    f: Callable[..., Tensor],
    argnums: Union[int, Sequence[int]] = 0,
    jit: bool = False,
    random_argnums: Optional[Sequence[int]] = None,
    shifts: Tuple[float, float] = (np.pi / 2, 2),
) -> Callable[..., Tensor]:
    """
    similar to `grad` function but using parameter shift internally instead of AD,
    vmap is utilized for evaluation, v2 also supports random generator for finite
    measurememt shot, only jax backend is supported, since no vmap randomness is
    available in tensorflow

    :param f: quantum function with weights in and expectation out
    :type f: Callable[..., Tensor]
    :param argnums: label which args should be differentiated,
        defaults to 0
    :type argnums: Union[int, Sequence[int]], optional
    :param jit: whether jit the original function `f` at the beginning,
        defaults to False
    :type jit: bool, optional
    :param shifts: two floats for the delta shift on the numerator and dominator,
        defaults to (pi/2, 2) for parameter shift
    :type shifts: Tuple[float, float]
    :return: the grad function
    :rtype: Callable[..., Tensor]
    """
    # TODO(@refraction-ray): replace with new status support for the sample API
    if jit is True:
        f = backend.jit(f)

    if isinstance(argnums, int):
        argnums = [argnums]

    if random_argnums is None:
        vfs = [backend.vmap(f, vectorized_argnums=i) for i in argnums]
    else:
        if isinstance(random_argnums, int):
            random_argnums = [random_argnums]
        vfs = [
            backend.vmap(f, vectorized_argnums=[i] + random_argnums) for i in argnums  # type: ignore
        ]

    def grad_f(*args: Any, **kws: Any) -> Any:
        grad_values = []
        for i in argnums:  # type: ignore
            shape = backend.shape_tuple(args[i])
            size = backend.sizen(args[i])
            onehot = backend.eye(size)
            onehot = backend.cast(onehot, args[i].dtype)
            onehot = backend.reshape(onehot, [size] + list(shape))
            onehot = shifts[0] * onehot
            nargs = list(args)
            arg = backend.reshape(args[i], [1] + list(shape))
            batched_arg = backend.tile(arg, [size] + [1 for _ in shape])
            nargs[i] = batched_arg + onehot
            nargs2 = list(args)
            nargs2[i] = batched_arg - onehot
            if random_argnums is not None:
                for j in random_argnums:
                    keys = []
                    key = args[j]
                    for _ in range(size):
                        key, subkey = backend.random_split(key)
                        keys.append(subkey)
                    nargs[j] = backend.stack(keys)
                    keys = []
                    for _ in range(size):
                        key, subkey = backend.random_split(key)
                        keys.append(subkey)
                    nargs2[j] = backend.stack(keys)
            r = (vfs[i](*nargs, **kws) - vfs[i](*nargs2, **kws)) / shifts[1]
            r = backend.reshape(r, shape)
            grad_values.append(r)
        if len(argnums) > 1:  # type: ignore
            return tuple(grad_values)
        return grad_values[0]

    return grad_f


def finite_difference_differentiator(
    f: Callable[..., Any],
    argnums: Tuple[int, ...] = (0,),
    shifts: Tuple[float, float] = (0.001, 0.002),
) -> Callable[..., Any]:
    # \bar{x}_j = \sum_i \bar{y}_i \frac{\Delta y_i}{\Delta x_j}
    # tf only now and designed for hardware, since we dont do batch evaluation
    import tensorflow as tf

    @tf.custom_gradient  # type: ignore
    def tf_function(*args: Any, **kwargs: Any) -> Any:
        y = f(*args, **kwargs)

        def grad(ybar: Any) -> Any:
            # only support one output
            delta_ms = []
            for argnum in argnums:
                delta_m = []
                xi = tf.reshape(args[argnum], [-1])
                xi_size = xi.shape[0]
                onehot = tf.one_hot(tf.range(xi_size), xi_size)
                for j in range(xi_size):
                    xi_plus = xi + tf.cast(shifts[0] * onehot[j], xi.dtype)
                    xi_minus = xi - tf.cast(shifts[0] * onehot[j], xi.dtype)
                    args_plus = list(args)
                    args_plus[argnum] = tf.reshape(xi_plus, args[argnum].shape)
                    args_minus = list(args)
                    args_minus[argnum] = tf.reshape(xi_minus, args[argnum].shape)
                    dy = f(*args_plus, **kwargs) - f(*args_minus, **kwargs)
                    dy /= shifts[-1]
                    delta_m.append(tf.reshape(dy, [-1]))
                delta_m = tf.stack(delta_m)
                delta_m = tf.transpose(delta_m)
                delta_ms.append(delta_m)
            dxs = [tf.zeros_like(arg) for arg in args]
            ybar_flatten = tf.reshape(ybar, [1, -1])
            for i, argnum in enumerate(argnums):
                dxs[argnum] = tf.cast(
                    tf.reshape(ybar_flatten @ delta_ms[i], args[argnum].shape),
                    args[argnum].dtype,
                )

            return tuple(dxs)

        return y, grad

    return tf_function  # type: ignore


def jax_jitted_function_save(filename: str, f: Callable[..., Any], *args: Any) -> None:
    """
    save a jitted jax function as a file

    :param filename: _description_
    :type filename: str
    :param f: the jitted function
    :type f: Callable[..., Any]
    :param args: example function arguments for ``f``
    """

    from jax import export  # type: ignore

    f_export = export.export(f)(*args)  # type: ignore
    barray = f_export.serialize()

    with open(filename, "wb") as file:
        file.write(barray)


jax_func_save = jax_jitted_function_save


def jax_jitted_function_load(filename: str) -> Callable[..., Any]:
    """
    load a jitted function from file

    :param filename: _description_
    :type filename: str
    :return: the loaded function
    :rtype: _type_
    """
    from jax import export  # type: ignore

    with open(filename, "rb") as f:
        barray = f.read()

    f_load = export.deserialize(barray)  # type: ignore

    return f_load.call  # type: ignore


jax_func_load = jax_jitted_function_load


PADDING_VALUE = -1
jaxlib: Any
ctg: Any
Mesh: Any
NamedSharding: Any
P: Any


def broadcast_py_object(obj: Any, shared_dir: Optional[str] = None) -> Any:
    """
    Broadcast a picklable Python object from process 0 to all other processes,
    with fallback mechanism from gRPC to file system based approach.

    This function first attempts to use gRPC-based broadcast. If that fails due to
    pickling issues, it falls back to a file system based approach that is more robust.

    :param obj: The Python object to broadcast. It must be picklable.
                This object should exist on process 0 and can be None on others.
    :type obj: Any
    :param shared_dir: Directory path for shared file system broadcast fallback.
                       If None, uses current directory. Only used in fallback mode.
    :type shared_dir: Optional[str], optional
    :return: The broadcasted object, now present on all processes.
    :rtype: Any
    """
    import jax
    from jax.experimental import multihost_utils

    try:
        result = broadcast_py_object_jax(obj)
        return result

    except pickle.UnpicklingError as e:
        # This block is executed if any process fails during the gRPC attempt.

        multihost_utils.sync_global_devices("grpc_broadcast_failed_fallback_sync")

        if jax.process_index() == 0:
            border = "=" * 80
            logger.warning(
                "\n%s\nJAX gRPC broadcast failed with error: %s\n"
                "--> Falling back to robust Shared File System broadcast method.\n%s",
                border,
                e,
                border,
            )

        return broadcast_py_object_fs(obj, shared_dir)


def broadcast_py_object_jax(obj: Any) -> Any:
    """
    Broadcast a picklable Python object from process 0 to all other processes
    within jax ditribution system.

    This function uses a two-step broadcast: first the size, then the data.
    This is necessary because `broadcast_one_to_all` requires the same
    shaped array on all hosts.

    :param obj: The Python object to broadcast. It must be picklable.
             This object should exist on process 0 and can be None on others.

    :return: The broadcasted object, now present on all processes.
    """
    import jax as jaxlib
    import pickle
    from jax.experimental import multihost_utils

    # Serialize to bytes on process 0, empty bytes on others
    if jaxlib.process_index() == 0:
        if obj is None:
            raise ValueError("Object to broadcast from process 0 cannot be None.")
        data = pickle.dumps(obj)
        logger.info(
            f"--- Size of object to be broadcast: {len(data) / 1024**2:.3f} MB ---"
        )

    else:
        data = b""

    # Step 1: Broadcast the length of the serialized data.
    # We send a single-element int32 array.
    length = np.array([len(data)], dtype=np.int32)
    length = multihost_utils.broadcast_one_to_all(length)

    length = int(length[0])  # type: ignore

    # Step 2: Broadcast the actual data.
    # Convert byte string to a uint8 array for broadcasting.
    send_arr_uint8 = np.frombuffer(data, dtype=np.uint8)
    padded_length = (length + 3) // 4 * 4
    if send_arr_uint8.size < padded_length:
        send_arr_uint8 = np.pad(  # type: ignore
            send_arr_uint8, (0, padded_length - send_arr_uint8.size), mode="constant"
        )
    send_arr_int32 = send_arr_uint8.astype(np.int32)
    # send_arr_int32 = jaxlib.numpy.array(send_arr_int32, dtype=np.int32)
    send_arr_int32 = jaxlib.device_put(send_arr_int32)

    jaxlib.experimental.multihost_utils.sync_global_devices("bulk_before")

    received_arr = multihost_utils.broadcast_one_to_all(send_arr_int32)

    received_arr = np.array(received_arr)
    received_arr_uint8 = received_arr.astype(np.uint8)

    # Step 3: Reconstruct the object from the received bytes.
    # Convert the NumPy array back to bytes, truncate any padding, and unpickle.
    received_data = received_arr_uint8[:length].tobytes()
    # if jaxlib.process_index() == 0:
    #     logger.info(f"Broadcasted object {obj}")
    return pickle.loads(received_data)


def broadcast_py_object_fs(
    obj: Any, shared_dir: Optional[str] = None, timeout_seconds: int = 300
) -> Any:
    """
    Broadcast a picklable Python object from process 0 to all other processes
    using a shared file system approach.

    This is a fallback method when gRPC-based broadcast fails. It uses UUID-based
    file communication to share objects between processes through a shared file system.

    :param obj: The Python object to broadcast. Must be picklable.
                Should exist on process 0, can be None on others.
    :type obj: Any
    :param shared_dir: Directory path for shared file system communication.
                       If None, uses current directory.
    :type shared_dir: Optional[str], optional
    :param timeout_seconds: Maximum time to wait for file operations before timing out.
                            Defaults to 300 seconds.
    :type timeout_seconds: int, optional
    :return: The broadcasted object, now present on all processes.
    :rtype: Any
    """
    # to_avoid very subtle bugs for broadcast tree_data on A800 clusters
    import jax
    from jax.experimental import multihost_utils

    if shared_dir is None:
        shared_dir = "."
    if jax.process_index() == 0:
        os.makedirs(shared_dir, exist_ok=True)

    id_comm_path = os.path.join(shared_dir, f".broadcast_temp_12318")
    transfer_id = ""

    if jax.process_index() == 0:
        transfer_id = str(uuid.uuid4())
        # print(f"[Process 0] Generated unique transfer ID: {transfer_id}", flush=True)
        with open(id_comm_path, "w") as f:
            f.write(transfer_id)

    multihost_utils.sync_global_devices("fs_broadcast_id_written")

    if jax.process_index() != 0:
        start_time = time.time()
        while not os.path.exists(id_comm_path):
            time.sleep(0.1)
            if time.time() - start_time > timeout_seconds:
                raise TimeoutError(
                    f"Process {jax.process_index()} timed out waiting for ID file: {id_comm_path}"
                )
        with open(id_comm_path, "r") as f:
            transfer_id = f.read()

    multihost_utils.sync_global_devices("fs_broadcast_id_read")
    if jax.process_index() == 0:
        try:
            os.remove(id_comm_path)
        except OSError:
            pass  # 如果文件已被其他进程快速清理，忽略错误

    # 定义本次传输使用的数据文件和标志文件路径
    data_path = os.path.join(shared_dir, f"{transfer_id}.data")
    done_path = os.path.join(shared_dir, f"{transfer_id}.done")

    result_obj = None

    if jax.process_index() == 0:
        if obj is None:
            raise ValueError("None cannot be broadcasted.")

        # print(f"[Process 0] Pickling object...", flush=True)
        pickled_data = pickle.dumps(obj)
        logger.info(
            f"[Process 0] Writing {len(pickled_data) / 1024**2:.3f} MB to {data_path}"
        )
        with open(data_path, "wb") as f:
            f.write(pickled_data)

        with open(done_path, "w") as f:
            pass
        logger.info(f"[Process 0] Write complete.")
        result_obj = obj
    else:
        # print(f"[Process {jax.process_index()}] Waiting for done file: {done_path}", flush=True)
        start_time = time.time()
        while not os.path.exists(done_path):
            time.sleep(0.1)
            if time.time() - start_time > timeout_seconds:
                raise TimeoutError(
                    f"Process {jax.process_index()} timed out waiting for done file: {done_path}"
                )

        # print(f"[Process {jax.process_index()}] Done file found. Reading data from {data_path}", flush=True)
        with open(data_path, "rb") as f:
            pickled_data = f.read()

        result_obj = pickle.loads(pickled_data)
        logger.info(f"[Process {jax.process_index()}] Object successfully loaded.")

    multihost_utils.sync_global_devices("fs_broadcast_read_complete")

    if jax.process_index() == 0:
        try:
            os.remove(data_path)
            os.remove(done_path)
            # print(f"[Process 0] Cleaned up temporary files for transfer {transfer_id}.", flush=True)
        except OSError as e:
            logger.info(
                f"[Process 0]: Failed to clean up temporary files: {e}",
            )

    return result_obj


class DistributedContractor:
    """
    A distributed tensor network contractor that parallelizes computations across multiple devices.

    This class uses cotengra to find optimal contraction paths and distributes the computational
    load across multiple devices (e.g., GPUs) for efficient tensor network calculations.
    Particularly useful for large-scale quantum circuit simulations and variational quantum algorithms.

    Example:
        >>> def nodes_fn(params):
        ...     c = tc.Circuit(4)
        ...     c.rx(0, theta=params[0])
        ...     return c.expectation_before([tc.gates.z(), [0]], reuse=False)
        >>> dc = DistributedContractor(nodes_fn, params)
        >>> value, grad = dc.value_and_grad(params)

    :param nodes_fn: Function that takes parameters and returns a list of tensor network nodes
    :type nodes_fn: Callable[[Tensor], List[Gate]]
    :param params: Initial parameters used to determine the tensor network structure
    :type params: Tensor
    :param cotengra_options: Configuration options passed to the cotengra optimizer. Defaults to None
    :type cotengra_options: Optional[Dict[str, Any]], optional
    :param devices: List of devices to use. If None, uses all available devices
    :type devices: Optional[List[Any]], optional
    :param mesh: Mesh object to use for distributed computation. If None, uses all available devices
    :type mesh: Optional[Any], optional
    """

    def __init__(
        self,
        nodes_fn: Callable[[Tensor], List[Gate]],
        params: Tensor,
        cotengra_options: Optional[Dict[str, Any]] = None,
        devices: Optional[List[Any]] = None,  # backward compatibility
        mesh: Optional[Any] = None,
        tree_data: Optional[Dict[str, Any]] = None,
    ) -> None:
        global jaxlib
        global ctg
        global Mesh
        global NamedSharding
        global P

        logger.info("Initializing DistributedContractor...")
        import cotengra as ctg
        from cotengra import ContractionTree
        import jax as jaxlib
        from jax.sharding import Mesh, NamedSharding, PartitionSpec as P

        self.nodes_fn = nodes_fn
        if mesh is not None:
            self.mesh = mesh
        elif devices is not None:
            self.mesh = Mesh(devices, axis_names=("devices",))
        else:
            self.mesh = Mesh(jaxlib.devices(), axis_names=("devices",))
        self.num_devices = len(self.mesh.devices)

        if self.num_devices <= 1:
            logger.info("DistributedContractor is running on a single device.")

        self._params_template = params
        self.params_sharding = jaxlib.tree_util.tree_map(
            lambda x: NamedSharding(self.mesh, P(*((None,) * x.ndim))),
            self._params_template,
        )
        self._backend = "jax"
        self._compiled_v_fns: Dict[
            Tuple[Callable[[Tensor], Tensor], str],
            Callable[[Any, Tensor, Tensor], Tensor],
        ] = {}
        self._compiled_vg_fns: Dict[
            Tuple[Callable[[Tensor], Tensor], str],
            Callable[[Any, Tensor, Tensor], Tensor],
        ] = {}

        logger.info("Running cotengra pathfinder... (This may take a while)")
        if tree_data is None:
            if params is None:
                raise ValueError("Please provide specific circuit parameters array.")
            if jaxlib.process_index() == 0:
                logger.info("Process 0: Running cotengra pathfinder...")
                tree_data = self._get_tree_data(
                    self.nodes_fn, self._params_template, cotengra_options  # type: ignore
                )

            # Step 2: Use the robust helper function to broadcast the tree object.
            # Process 0 sends its computed `tree_object`.
            # Other processes send `None`, but receive the object from process 0.

            if jaxlib.process_count() > 1:
                # self.tree = broadcast_py_object(tree_object)
                jaxlib.experimental.multihost_utils.sync_global_devices("tree_before")
                logger.info(
                    f"Process {jaxlib.process_index()}: Synchronizing contraction path..."
                )
                tree_data = broadcast_py_object(tree_data)
                jaxlib.experimental.multihost_utils.sync_global_devices("tree_after")
        else:
            logger.info("Using pre-computed contraction path.")
        if tree_data is None:
            raise ValueError("Contraction path data is missing.")

        self.tree = ContractionTree.from_path(
            inputs=tree_data["inputs"],
            output=tree_data["output"],
            size_dict=tree_data["size_dict"],
            path=tree_data["path"],
        )

        # Restore slicing information
        for ind, _ in tree_data["sliced_inds"].items():
            self.tree.remove_ind_(ind)

        logger.info(
            f"Process {jaxlib.process_index()}: Contraction path successfully synchronized."
        )
        actual_num_slices = self.tree.nslices

        self._report_tree_info()

        slices_per_device = int(np.ceil(actual_num_slices / self.num_devices))
        padded_size = slices_per_device * self.num_devices
        slice_indices = np.arange(actual_num_slices)
        padded_slice_indices = np.full(padded_size, PADDING_VALUE, dtype=np.int32)
        padded_slice_indices[:actual_num_slices] = slice_indices

        # Reshape for distribution and define the sharding rule
        batched_indices = padded_slice_indices.reshape(
            self.num_devices, slices_per_device
        )
        # Sharding rule: split the first axis (the one for devices) across the 'devices' mesh axis
        self.sharding = NamedSharding(self.mesh, P("devices", None))
        # Place the tensor on devices according to the rule
        self.batched_slice_indices = jaxlib.device_put(batched_indices, self.sharding)

        # self.batched_slice_indices = backend.convert_to_tensor(
        #     padded_slice_indices.reshape(self.num_devices, slices_per_device)
        # )
        print(
            f"Distributing across {self.num_devices} devices. Each device will sequentially process "
            f"up to {slices_per_device} slices."
        )

        self._compiled_vg_fn = None
        self._compiled_v_fn = None

        logger.info("Initialization complete.")

    def _report_tree_info(self) -> None:
        print("\n--- Contraction Path Info ---")
        actual_num_slices = self.tree.nslices
        stats = self.tree.contract_stats()
        print(f"Path found with {actual_num_slices} slices.")
        print(
            f"Arithmetic Intensity (higher is better): {self.tree.arithmetic_intensity():.2f}"
        )
        print("flops (TFlops):", stats["flops"] / 2**40 / self.num_devices)
        print("write (GB):", stats["write"] / 2**27 / actual_num_slices)
        print("size (GB):", stats["size"] / 2**27)
        print("-----------------------------\n")

    @staticmethod
    def _get_tree_data(
        nodes_fn: Callable[[Tensor], List[Gate]],
        params: Tensor,
        cotengra_options: Optional[Dict[str, Any]] = None,
    ) -> Dict[str, Any]:
        global ctg

        import cotengra as ctg

        local_cotengra_options = (cotengra_options or {}).copy()

        nodes = nodes_fn(params)
        tn_info, _ = get_tn_info(nodes)
        default_cotengra_options = {
            "slicing_reconf_opts": {"target_size": 2**28},
            "max_repeats": 128,
            "minimize": "write",
            "parallel": "auto",
            "progbar": True,
        }
        default_cotengra_options.update(local_cotengra_options)

        opt = ctg.ReusableHyperOptimizer(**default_cotengra_options)
        tree_object = opt.search(*tn_info)
        tree_data = {
            "inputs": tree_object.inputs,
            "output": tree_object.output,
            "size_dict": tree_object.size_dict,
            "path": tree_object.get_path(),
            "sliced_inds": tree_object.sliced_inds,
        }
        return tree_data

    @staticmethod
    def find_path(
        nodes_fn: Callable[[Tensor], Tensor],
        params: Tensor,
        cotengra_options: Optional[Dict[str, Any]] = None,
        filepath: Optional[str] = None,
    ) -> None:
        tree_data = DistributedContractor._get_tree_data(
            nodes_fn, params, cotengra_options
        )
        if filepath is not None:
            with open(filepath, "wb") as f:
                pickle.dump(tree_data, f)
            logger.info(f"Contraction path data successfully saved to '{filepath}'.")

    @classmethod
    def from_path(
        cls,
        filepath: str,
        nodes_fn: Callable[[Tensor], List[Gate]],
        devices: Optional[List[Any]] = None,  # backward compatibility
        mesh: Optional[Any] = None,
        params: Any = None,
    ) -> "DistributedContractor":
        with open(filepath, "rb") as f:
            tree_data = pickle.load(f)

        # Each process loads the file independently. No broadcast is needed.
        # We pass the loaded `tree_data` directly to __init__ to trigger the second workflow.
        return cls(
            nodes_fn=nodes_fn,
            params=params,
            mesh=mesh,
            devices=devices,
            tree_data=tree_data,
        )

    def _get_single_slice_contraction_fn(
        self, op: Optional[Callable[[Tensor], Tensor]] = None
    ) -> Callable[[Any, Tensor, int], Tensor]:
        if op is None:
            op = backend.sum

        def single_slice_contraction(
            tree: ctg.ContractionTree, params: Tensor, slice_idx: int
        ) -> Tensor:
            nodes = self.nodes_fn(params)
            _, standardized_nodes = get_tn_info(nodes)
            input_arrays = [node.tensor for node in standardized_nodes]
            sliced_arrays = tree.slice_arrays(input_arrays, slice_idx)
            result = tree.contract_core(sliced_arrays, backend=self._backend)
            return op(result)

        return single_slice_contraction

    def _get_device_sum_vg_fn(
        self,
        op: Optional[Callable[[Tensor], Tensor]] = None,
        output_dtype: Optional[str] = None,
    ) -> Callable[[Any, Tensor, Tensor], Tuple[Tensor, Tensor]]:
        post_processing = lambda x: backend.real(backend.sum(x))
        if op is None:
            op = post_processing
        base_fn = self._get_single_slice_contraction_fn(op=op)
        # to ensure the output is real so that can be differentiated
        single_slice_vg_fn = jaxlib.value_and_grad(base_fn, argnums=1)

        if output_dtype is None:
            output_dtype = rdtypestr

        def device_sum_fn(
            tree: ctg.ContractionTree, params: Tensor, slice_indices_for_device: Tensor
        ) -> Tuple[Tensor, Tensor]:
            def scan_body(
                carry: Tuple[Tensor, Tensor], slice_idx: Tensor
            ) -> Tuple[Tuple[Tensor, Tensor], None]:
                acc_value, acc_grads = carry

                def compute_and_add() -> Tuple[Tensor, Tensor]:
                    value_slice, grads_slice = single_slice_vg_fn(
                        tree, params, slice_idx
                    )
                    new_value = acc_value + value_slice
                    new_grads = jaxlib.tree_util.tree_map(
                        jaxlib.numpy.add, acc_grads, grads_slice
                    )
                    return new_value, new_grads

                def do_nothing() -> Tuple[Tensor, Tensor]:
                    return acc_value, acc_grads

                return (
                    jaxlib.lax.cond(
                        slice_idx == PADDING_VALUE, do_nothing, compute_and_add
                    ),
                    None,
                )

            initial_carry = (
                backend.cast(backend.convert_to_tensor(0.0), dtype=output_dtype),
                jaxlib.tree_util.tree_map(lambda x: jaxlib.numpy.zeros_like(x), params),
            )
            (final_value, final_grads), _ = jaxlib.lax.scan(
                scan_body, initial_carry, slice_indices_for_device
            )
            return final_value, final_grads

        return device_sum_fn

    def _get_device_sum_v_fn(
        self,
        op: Optional[Callable[[Tensor], Tensor]] = None,
        output_dtype: Optional[str] = None,
    ) -> Callable[[Any, Tensor, Tensor], Tensor]:
        base_fn = self._get_single_slice_contraction_fn(op=op)
        if output_dtype is None:
            output_dtype = dtypestr

        def device_sum_fn(
            tree: ctg.ContractionTree, params: Tensor, slice_indices_for_device: Tensor
        ) -> Tensor:
            def scan_body(
                carry_value: Tensor, slice_idx: Tensor
            ) -> Tuple[Tensor, None]:
                def compute_and_add() -> Tensor:
                    return carry_value + base_fn(tree, params, slice_idx)

                return (
                    jaxlib.lax.cond(
                        slice_idx == PADDING_VALUE, lambda: carry_value, compute_and_add
                    ),
                    None,
                )

            initial_carry = backend.cast(
                backend.convert_to_tensor(0.0), dtype=output_dtype
            )
            final_value, _ = jaxlib.lax.scan(
                scan_body, initial_carry, slice_indices_for_device
            )
            return final_value

        return device_sum_fn

    def _get_or_compile_fn(
        self,
        cache: Dict[
            Tuple[Callable[[Tensor], Tensor], str],
            Callable[[Any, Tensor, Tensor], Tensor],
        ],
        fn_getter: Callable[..., Any],
        op: Optional[Callable[[Tensor], Tensor]],
        output_dtype: Optional[str],
        is_grad_fn: bool,
    ) -> Callable[[Any, Tensor, Tensor], Tensor]:
        """
        Gets a compiled pmap-ed function from cache or compiles and caches it.

        The cache key is a tuple of (op, output_dtype). Caution on lambda function!

        Returns:
            The compiled, pmap-ed JAX function.
        """
        cache_key = (op, output_dtype)
        if cache_key not in cache:
            device_fn = fn_getter(op=op, output_dtype=output_dtype)

            def global_aggregated_fn(
                tree: Any, params: Any, batched_slice_indices: Tensor
            ) -> Any:
                # Use jax.vmap to apply the per-device function across the sharded data.
                # vmap maps `device_fn` over the first axis (0) of `batched_slice_indices`.
                # `tree` and `params` are broadcasted (in_axes=None) to each call.
                vmapped_device_fn = jaxlib.vmap(
                    device_fn, in_axes=(None, None, 0), out_axes=0
                )
                device_results = vmapped_device_fn(tree, params, batched_slice_indices)

                # Now, `device_results` is a sharded PyTree (one result per device).
                # We aggregate them using jnp.sum, which JAX automatically compiles
                # into a cross-device AllReduce operation.

                if is_grad_fn:
                    # `device_results` is a (value, grad) tuple of sharded arrays
                    device_values, device_grads = device_results

                    # Replace psum with jnp.sum
                    global_value = jaxlib.numpy.sum(device_values, axis=0)
                    global_grad = jaxlib.tree_util.tree_map(
                        lambda g: jaxlib.numpy.sum(g, axis=0), device_grads
                    )
                    return global_value, global_grad
                else:
                    # `device_results` is just the sharded values
                    return jaxlib.numpy.sum(device_results, axis=0)

            #  Compile the global function with jax.jit and specify shardings.
            # `params` are replicated (available everywhere).

            in_shardings = (self.params_sharding, self.sharding)

            if is_grad_fn:
                # Returns (value, grad), so out_sharding must be a 2-tuple.
                # `value` is a replicated scalar -> P()
                sharding_for_value = NamedSharding(self.mesh, P())
                # `grad` is a replicated PyTree with the same structure as params.
                sharding_for_grad = self.params_sharding
                out_shardings = (sharding_for_value, sharding_for_grad)
            else:
                # Returns a single scalar value -> P()
                out_shardings = NamedSharding(self.mesh, P())

            compiled_fn = jaxlib.jit(
                global_aggregated_fn,
                # `tree` is a static argument, its value is compiled into the function.
                static_argnums=(0,),
                # Specify how inputs are sharded.
                in_shardings=in_shardings,
                # Specify how the output should be sharded.
                out_shardings=out_shardings,
            )
            cache[cache_key] = compiled_fn  # type: ignore
        return cache[cache_key]  # type: ignore

    def value_and_grad(
        self,
        params: Tensor,
        # aggregate: bool = True,
        op: Optional[Callable[[Tensor], Tensor]] = None,
        output_dtype: Optional[str] = None,
    ) -> Tuple[Tensor, Tensor]:
        """
        Calculates the value and gradient, compiling the pmap function if needed for the first call.

        :param params: Parameters for the `nodes_fn` input
        :type params: Tensor
        :param op: Optional post-processing function for the output, defaults to None (corresponding to `backend.real`)
            op is a cache key, so dont directly pass lambda function for op
        :type op: Optional[Callable[[Tensor], Tensor]], optional
        :param output_dtype: dtype str for the output of `nodes_fn`, defaults to None (corresponding to `rdtypestr`)
        :type output_dtype: Optional[str], optional
        """
        compiled_vg_fn = self._get_or_compile_fn(
            cache=self._compiled_vg_fns,
            fn_getter=self._get_device_sum_vg_fn,
            op=op,
            output_dtype=output_dtype,
            is_grad_fn=True,
        )

        total_value, total_grad = compiled_vg_fn(
            self.tree, params, self.batched_slice_indices
        )
        return total_value, total_grad

    def value(
        self,
        params: Tensor,
        # aggregate: bool = True,
        op: Optional[Callable[[Tensor], Tensor]] = None,
        output_dtype: Optional[str] = None,
    ) -> Tensor:
        """
        Calculates the value, compiling the pmap function for the first call.

        :param params: Parameters for the `nodes_fn` input
        :type params: Tensor
        :param op: Optional post-processing function for the output, defaults to None (corresponding to identity)
            op is a cache key, so dont directly pass lambda function for op
        :type op: Optional[Callable[[Tensor], Tensor]], optional
        :param output_dtype: dtype str for the output of `nodes_fn`, defaults to None (corresponding to `dtypestr`)
        :type output_dtype: Optional[str], optional
        """
        compiled_v_fn = self._get_or_compile_fn(
            cache=self._compiled_v_fns,
            fn_getter=self._get_device_sum_v_fn,
            op=op,
            output_dtype=output_dtype,
            is_grad_fn=False,
        )

        total_value = compiled_v_fn(self.tree, params, self.batched_slice_indices)
        return total_value

    def grad(
        self,
        params: Tensor,
        op: Optional[Callable[[Tensor], Tensor]] = None,
        output_dtype: Optional[str] = None,
    ) -> Tensor:
        _, grad = self.value_and_grad(params, op=op, output_dtype=output_dtype)
        return grad



"""
DQAS application kernels as vag functions
"""

from functools import lru_cache, partial, reduce
import logging
import operator
from typing import (
    List,
    Sequence,
    Any,
    Tuple,
    Callable,
    Iterator,
    Optional,
    Union,
)

import networkx as nx
import numpy as np
import tensorflow as tf

from ..gates import array_to_tensor
from .. import cons
from .. import gates as G
from ..circuit import Circuit
from ..densitymatrix import DMCircuit

from .layers import generate_qubits
from .dqas import get_op_pool

logger = logging.getLogger(__name__)

try:
    import cirq
    import sympy
    import tensorflow_quantum as tfq

except ImportError as e:
    logger.warning(e)
    logger.warning("Therefore some functionality in %s may not work" % __name__)


Array = Any  # np.array
Opt = Any  # tf.keras.optimizer
Model = Any  # tf.keras.Model
Tensor = Any
Graph = Any


## GHZ circuit application


def GHZ_vag(
    gdata: Any, nnp: Tensor, preset: Sequence[int], verbose: bool = False, n: int = 3
) -> Tuple[Tensor, Tensor]:
    # gdata = None
    reference_state = np.zeros([2**n])
    #     W states benchmarks
    #     for i in range(n):
    #         reference_state[2**(i)] = 1/np.sqrt(n)
    reference_state[0] = 1 / np.sqrt(2)
    reference_state[-1] = 1 / np.sqrt(2)
    reference_state = tf.constant(reference_state, dtype=tf.complex64)
    nnp = nnp.numpy()  # real
    pnnp = [nnp[i, j] for i, j in enumerate(preset)]
    pnnp = array_to_tensor(np.array(pnnp))  # complex
    circuit = Circuit(n)
    cset = get_op_pool()

    with tf.GradientTape() as t:
        t.watch(pnnp)
        for i, j in enumerate(preset):
            gate = cset[j]
            if gate[0].startswith("r"):
                getattr(circuit, gate[0])(gate[1], theta=pnnp[i])
            elif len(gate[0]) == 1:
                getattr(circuit, gate[0])(gate[1])
            elif gate[0] == "CNOT":
                circuit.CNOT(gate[1], gate[2])  # type: ignore
        s = circuit.wavefunction()
        s = tf.reshape(
            s,
            [2**n],
        )
        loss = tf.math.reduce_sum(
            tf.math.abs(s - reference_state)
        )  # better for overlap objective to optimize
    #   loss = (tf.math.abs(tf.tensordot(s, reference_state, [0,0]))-1.)**2
    if verbose:
        print(s.numpy())
    gr = t.gradient(loss, pnnp)
    if gr is None:
        # if all gates in preset are not trainable, then gr returns None instead of 0s
        gr = tf.zeros_like(pnnp)
    gr = cons.backend.real(gr)
    gr = tf.where(tf.math.is_nan(gr), 0.0, gr)
    gmatrix = np.zeros_like(nnp)
    for i, j in enumerate(preset):
        gmatrix[i, j] = gr[i]
    gmatrix = tf.constant(gmatrix)
    return loss, gmatrix


## QAOA application


@lru_cache()
def energy(i: int, n: int, g: Graph) -> float:
    """
    maxcut energy for n qubit wavefunction i-th basis

    :param i: ranged from 0 to 2**n-1
    :param n: number of qubits
    :param g: nx.Graph
    :return:
    """
    basis = bin(i)[2:].zfill(n)
    r = 0
    for e in g.edges:
        r += g[e[0]][e[1]].get("weight", 1.0) * int(basis[e[0]] != basis[e[1]])
    return r


def ave_func(
    state: Array,
    g: Graph,
    *fs: Union[
        Tuple[Callable[[float], float], Callable[[Tensor], Tensor]],
        Tuple[
            Callable[[float], float],
            Callable[[Tensor], Tensor],
            Callable[[Tensor, Tensor], Tensor],
        ],
    ],
) -> Sequence[Tensor]:
    """

    :param state: 1D array for full wavefunction, the basis is in lexcical order
    :param g: nx.Graph
    :param fs: transformation functions before averaged
    :return:
    """
    n = int(np.log2(len(state)))
    ebasis = [energy(i, n, g) for i in range(len(state))]
    result = []
    for ftuple in fs:
        if len(ftuple) == 2:
            f, f2 = ftuple  # type: ignore
        else:
            f, f2, f3 = ftuple  # type: ignore
        r = [f(e) for e in ebasis]
        if len(ftuple) == 3:
            r = f3(r, cons.backend.abs(state) ** 2)
        r = array_to_tensor(np.array(r))

        result.append(
            f2(
                cons.backend.real(
                    cons.backend.tensordot(
                        r,
                        cons.backend.cast(
                            cons.backend.abs(state) ** 2, dtype="complex64"
                        ),
                        [0, 0],
                    )
                )
            )
        )
    return result


def exp_forward(
    theta: Tensor,
    preset: Sequence[int],
    g: Graph,
    *fs: Tuple[Callable[[float], float], Callable[[Tensor], Tensor]],
) -> Sequence[Tensor]:
    n = len(g.nodes)
    ci = Circuit(n)
    cset = get_op_pool()
    for i, j in enumerate(preset):
        if callable(cset[j]):
            cset[j](ci, theta[i], g)
        else:
            layer, graph = cset[j]
            layer(ci, theta[i], graph)

    state = ci.wavefunction()
    losses = ave_func(state, g, *fs)
    return losses


def _identity(s: Any) -> Any:
    return s


def _neg(s: Any) -> Any:
    return -s


def _exp_fun(s: Any, lbd: float = 1.0) -> Tensor:
    return cons.backend.exp(-lbd * s)


def _overlap_fun(s: Any, overlap_threhold: float = 0.0) -> Tensor:
    if s >= overlap_threhold > 0:
        return 1.0
    return 0.0


def cvar(r: List[float], p: Tensor, percent: float = 0.2) -> Sequence[float]:
    """
    as f3

    :param r:
    :param p:
    :param percent:
    :return:
    """

    rs = sorted(
        [(i, j) for i, j in enumerate(r)], key=lambda s: -s[1]
    )  # larger to smaller
    sump = 0.0
    count = 0
    while sump < percent:
        if sump + p[rs[count][0]] > percent:
            r[rs[count][0]] = (percent - sump) / (p[rs[count][0]]) * r[rs[count][0]]
            count += 1
            break
        else:
            sump += p[rs[count][0]]
            count += 1

    for i in range(count, len(r)):
        r[rs[i][0]] = 0
    r = [k / percent for k in r]
    return r


def qaoa_vag(
    gdata: Graph,
    nnp: Tensor,
    preset: Sequence[int],
    f: Optional[Tuple[Callable[[float], float], Callable[[Tensor], Tensor]]] = None,
    forward_func: Optional[
        Callable[
            [
                Tensor,
                Sequence[int],
                Graph,
                Tuple[Callable[[float], float], Callable[[Tensor], Tensor]],
            ],
            Tuple[Tensor, Tensor],
        ]
    ] = None,
    verbose_fs: Optional[
        Sequence[Tuple[Callable[[float], float], Callable[[Tensor], Tensor]]]
    ] = None,
) -> Tuple[Tensor, Tensor]:
    if forward_func is None:
        forward_func = exp_forward  # type: ignore
    if f is None:
        f = (_identity, _neg)
    nnp = nnp.numpy()  # real
    pnnp = [nnp[i, j] for i, j in enumerate(preset)]
    pnnp = array_to_tensor(np.array(pnnp))  # complex
    with tf.GradientTape() as t:
        t.watch(pnnp)
        loss = forward_func(pnnp, preset, gdata, f)  # type: ignore
    gr = t.gradient(loss, pnnp)
    if verbose_fs:
        for vf in verbose_fs:
            print(forward_func(pnnp, preset, gdata, vf))  # type: ignore
    if gr is None:
        # if all gates in preset are not trainable, then gr returns None instead of 0s
        gr = tf.zeros_like(pnnp)
    gr = cons.backend.real(gr)
    gr = tf.where(tf.math.is_nan(gr), 0.0, gr)
    gmatrix = np.zeros_like(nnp)
    for i, j in enumerate(preset):
        gmatrix[i, j] = gr[i]
    gmatrix = tf.constant(gmatrix)
    return loss[0], gmatrix


def qaoa_block_vag(
    gdata: Graph,
    nnp: Tensor,
    preset: Sequence[int],
    f: Tuple[Callable[[float], float], Callable[[Tensor], Tensor]],
) -> Tuple[Tensor, Tensor]:
    """
    QAOA block encoding kernel, support 2 params in one op

    :param gdata:
    :param nnp:
    :param preset:
    :param f:
    :return:
    """
    # for universality, nnp always occupy two rows for each block
    # for a more general approach of multi params in one op, see ``quantum_mp_qaoa_vag``

    nnp = nnp.numpy()  # real
    pnnp = []
    ops = get_op_pool()
    for i, j in enumerate(preset):
        # print(ops[j].__repr__)
        if ops[j].__repr__.endswith("_block"):
            pnnp.append([nnp[2 * i, j], nnp[2 * i + 1, j]])
        else:
            pnnp.append([nnp[2 * i, j]])
    # pnnp = array_to_tensor(np.array(pnnp))  # complex
    pnnp = tf.ragged.constant(pnnp, dtype=getattr(tf, cons.dtypestr))
    with tf.GradientTape() as t:
        t.watch(pnnp.values)  # type: ignore
        loss = exp_forward(pnnp, preset, gdata, f)
    gr = t.gradient(loss, pnnp.values)  # type: ignore
    if gr is None:
        # if all gates in preset are not trainable, then gr returns None instead of 0s
        gr = tf.zeros_like(pnnp)
    else:
        gr = cons.backend.real(gr)
        gr = tf.where(tf.math.is_nan(gr), 0.0, gr)
        # if all gates in preset are not trainable, then gr returns None instead of 0s
        gr = pnnp.with_values(gr)  # type: ignore

    gr = cons.backend.real(gr)

    gmatrix = np.zeros_like(nnp)
    for j in range(gr.shape[0]):
        if gr[j].shape[0] == 2:
            gmatrix[2 * j, preset[j]] = gr[j][0]
            gmatrix[2 * j + 1, preset[j]] = gr[j][1]
        else:  # 1
            gmatrix[2 * j, preset[j]] = gr[j][0]

    gmatrix = tf.constant(gmatrix)
    return loss[0], gmatrix


qaoa_vag_energy = partial(qaoa_vag, f=(_identity, _neg))
qaoa_block_vag_energy = partial(qaoa_block_vag, f=(_identity, _neg))


def evaluate_vag(
    params: Array,
    preset: Sequence[int],
    g: Graph,
    lbd: float = 0.0,
    overlap_threhold: float = 0.0,
) -> Tuple[Tensor, Tensor, Tensor, Tensor]:
    """
    value and gradient, currently only tensorflow backend is supported
    jax and numpy seems to be slow in circuit simulation anyhow.
    *deprecated*

    :param params:
    :param preset:
    :param g:
    :param lbd: if lbd=0, take energy as objective
    :param overlap_threhold: if as default 0, overlap will not compute in the process
    :return:
    """
    params = array_to_tensor(params)  # complexify params
    _exp_fun_partial = partial(_exp_fun, lbd=lbd)
    _overlap_fun_partial = partial(_overlap_fun, overlap_threhold=overlap_threhold)
    with tf.GradientTape() as t:
        t.watch(params)
        expe, ene, probasum = exp_forward(
            params,
            preset,
            g,
            (_exp_fun_partial, cons.backend.log),  # gibbs objective
            (_identity, _neg),  # energy
            (_overlap_fun_partial, _identity),  # probability
        )

    if lbd == 0:
        gr = t.gradient(ene, params)
    else:
        gr = t.gradient(expe, params)
    return expe, ene, cons.backend.real(gr), probasum


## noise qaoa tensorcircuits setup


def noise_forward(
    theta: Tensor,
    preset: Sequence[int],
    g: Graph,
    measure_func: Callable[[DMCircuit, Graph], Tensor],
    is_mc: bool = False,
) -> Tensor:
    n = len(g.nodes)
    if is_mc is True:
        ci = Circuit(n)
    else:
        ci = DMCircuit(n)  # type: ignore
    cset = get_op_pool()
    for i, j in enumerate(preset):
        if len(cset[j]) == 3:  # (noiselayer, graph, [0.2])
            layer, graph, params = cset[j]
            layer(ci, theta[i], graph, *params)
        elif len(cset[j]) == 4:  # (rxlayer, graph, noiselayer, [0.1])
            layer, graph, noisemodel, params = cset[j]
            layer(ci, theta[i], graph)
            noisemodel(ci, g, *params)
        elif len(cset[j]) == 2:  # (noiselayer, [0.2]) no noise
            layer, params = cset[j]
            layer(ci, theta[i], g, *params)
        else:  # len == 1
            cset[j][0](ci, theta[i], g)

    loss = measure_func(ci, g)  # type: ignore
    return loss


def maxcut_measurements_tc(c: Union[Circuit, DMCircuit], g: Graph) -> Tensor:
    loss = 0.0
    for e in g.edges:
        loss += (
            g[e[0]][e[1]]["weight"]
            * 0.5
            * (c.expectation((G.z(), [e[0]]), (G.z(), [e[1]])) - 1)  # type: ignore
        )
    return loss


def tfim_measurements_tc(
    c: Union[Circuit, DMCircuit],
    g: Graph,
    hzz: float = 1.0,
    hx: float = 0.0,
    hz: float = 0.0,
) -> Tensor:
    loss = 0.0
    for e in g.edges:
        loss += (
            g[e[0]][e[1]]["weight"]
            * hzz
            * c.expectation((G.z(), [e[0]]), (G.z(), [e[1]]))  # type: ignore
        )
    if hx != 0.0:
        for i in range(len(g.nodes)):
            loss += hx * c.expectation((G.x(), [i]))  # type: ignore
    if hz != 0.0:
        for i in range(len(g.nodes)):
            loss += hz * c.expectation((G.z(), [i]))  # type: ignore
    return loss


def heisenberg_measurements_tc(
    c: Union[Circuit, DMCircuit],
    g: Graph,
    hzz: float = 1.0,
    hxx: float = 1.0,
    hyy: float = 1.0,
    hz: float = 0.0,
    hx: float = 0.0,
    hy: float = 0.0,
    reuse: bool = True,
) -> Tensor:
    loss = 0.0
    for e in g.edges:
        loss += (
            g[e[0]][e[1]]["weight"]
            * hzz
            * c.expectation((G.z(), [e[0]]), (G.z(), [e[1]]), reuse=reuse)  # type: ignore
        )
        loss += (
            g[e[0]][e[1]]["weight"]
            * hyy
            * c.expectation((G.y(), [e[0]]), (G.y(), [e[1]]), reuse=reuse)  # type: ignore
        )
        loss += (
            g[e[0]][e[1]]["weight"]
            * hxx
            * c.expectation((G.x(), [e[0]]), (G.x(), [e[1]]), reuse=reuse)  # type: ignore
        )
    if hx != 0:
        for i in range(len(g.nodes)):
            loss += hx * c.expectation((G.x(), [i]), reuse=reuse)  # type: ignore
    if hy != 0:
        for i in range(len(g.nodes)):
            loss += hy * c.expectation((G.y(), [i]), reuse=reuse)  # type: ignore
    if hz != 0:
        for i in range(len(g.nodes)):
            loss += hz * c.expectation((G.z(), [i]), reuse=reuse)  # type: ignore
    return loss


def qaoa_noise_vag(
    gdata: Graph,
    nnp: Tensor,
    preset: Sequence[int],
    measure_func: Optional[Callable[[DMCircuit, Graph], Tensor]] = None,
    forward_func: Optional[
        Callable[
            [Tensor, Sequence[int], Graph, Callable[[DMCircuit, Graph], Tensor]],
            Tensor,
        ]
    ] = None,
    **kws: Any,
) -> Tuple[Tensor, Tensor]:
    if measure_func is None:
        measure_func = maxcut_measurements_tc
    if forward_func is None:
        forward_func = noise_forward
    nnp = nnp.numpy()  # real
    pnnp = [nnp[i, j] for i, j in enumerate(preset)]
    pnnp = array_to_tensor(np.array(pnnp))  # complex
    with tf.GradientTape() as t:
        t.watch(pnnp)
        loss = forward_func(pnnp, preset, gdata, measure_func, **kws)
        loss = cons.backend.real(loss)
    gr = t.gradient(loss, pnnp)
    if gr is None:
        # if all gates in preset are not trainable, then gr returns None instead of 0s
        gr = tf.zeros_like(pnnp)
    gr = cons.backend.real(gr)
    gr = tf.where(tf.math.is_nan(gr), 0.0, gr)
    gmatrix = np.zeros_like(nnp)
    for i, j in enumerate(preset):
        gmatrix[i, j] = gr[i]
    gmatrix = tf.constant(gmatrix)
    # print("loss", loss, "g\n", gmatrix)
    return loss, gmatrix


def qaoa_train(
    preset: Sequence[int],
    g: Union[Graph, Iterator[Graph]],
    *,
    epochs: int = 100,
    batch: int = 1,
    initial_param: Optional[Array] = None,
    opt: Any = None,
    lbd: float = 0.0,
    overlap_threhold: float = 0.0,
    verbose: bool = True,
) -> Tuple[Array, Sequence[float], Sequence[float], Sequence[float]]:
    """
    training QAOA with only optimizing circuit parameters, can be well replaced with more general function `DQAS_search`

    :param preset:
    :param g:
    :param epochs:
    :param batch:
    :param initial_param:
    :param opt:
    :param lbd:
    :param overlap_threhold:
    :param verbose:
    :return:
    """
    ## depracated qaoa train

    if initial_param is None:
        initial_param = np.random.normal(loc=0.3, scale=0.05, size=[len(preset)])
    theta = tf.Variable(initial_value=initial_param, dtype=tf.float32)
    # this initialization is near optimal as this parallels QA protocol
    if opt is None:
        opt = tf.keras.optimizers.Adam(learning_rate=0.02)
    gibbs_history = []
    mean_history = []
    overlap_history = []
    if isinstance(g, nx.Graph):

        def one_generator() -> Iterator[Graph]:
            while True:
                yield g

        gi = one_generator()
    else:
        gi = g
    for i in range(epochs):
        means = []
        grs = []
        overlaps = []
        gibbs = []
        for _ in range(batch):
            gdata = gi.send(None)  # type: ignore
            value, ene, gra, probasum = evaluate_vag(
                theta.numpy(),
                preset,
                gdata,
                lbd=lbd,
                overlap_threhold=overlap_threhold,
            )

            gibbs.append(value.numpy())
            means.append(ene.numpy())
            overlaps.append(probasum.numpy())
            grs.append(gra.numpy())
        overlap_history.append(np.mean(overlaps))
        gibbs_history.append(np.mean(gibbs))
        mean_history.append(np.mean(means))
        gr = tf.reduce_mean(tf.constant(grs), axis=0)
        opt.apply_gradients([(gr, theta)])
        if verbose:
            print("epoch:", i)
            print("Gibbs objective:", gibbs_history[-1])
            print("mean energy:", mean_history[-1])
            print("overlap:", overlap_history[-1])
            print("trainable weights:", theta.numpy())
    return theta.numpy(), mean_history, gibbs_history, overlap_history


def compose_tc_circuit_with_multiple_pools(
    theta: Tensor,
    preset: Sequence[int],
    g: Graph,
    pool_choice: Sequence[int],
    cset: Optional[Sequence[Any]] = None,
    measure_func: Optional[Callable[[DMCircuit, Graph], Tensor]] = None,
) -> Circuit:
    n = len(g.nodes)
    ci = Circuit(n)
    if cset is None:
        cset = get_op_pool()
    mem = 0
    for i, j in enumerate(preset):
        ele = cset[pool_choice[i]][j]
        if isinstance(ele, tuple) or isinstance(ele, list):
            gate = ele[0]
            index = ele[1]
        else:
            index = [mem % n]
            gate = ele
            mem += 1
        if gate.lower() in ["cnot"]:
            getattr(ci, gate)(*index)
        else:
            getattr(ci, gate)(*index, theta=theta[i])
    return ci


def gatewise_vqe_vag(
    gdata: Graph,
    nnp: Tensor,
    preset: Sequence[int],
    pool_choice: Sequence[int],
    measure_func: Optional[Callable[[Union[Circuit, DMCircuit], Graph], Tensor]] = None,
) -> Tuple[Tensor, Tensor]:
    nnp = nnp.numpy()  # real
    pnnp = []
    cset = get_op_pool()
    if measure_func is None:
        measure_func = maxcut_measurements_tc
    for i, j in enumerate(preset):
        k = pool_choice[i]
        if j >= len(cset[k]):
            j = len(cset[k]) - 1
            preset[i] = j  # type: ignore
        pnnp.append(nnp[i, j])
    pnnp = array_to_tensor(np.array(pnnp))  # complex
    with tf.GradientTape() as tape:
        tape.watch(pnnp)
        ci = compose_tc_circuit_with_multiple_pools(pnnp, preset, gdata, pool_choice)
        loss = measure_func(ci, gdata)
        loss = cons.backend.real(loss)
    gr = tape.gradient(loss, pnnp)
    if gr is None:
        # if all gates in preset are not trainable, then gr returns None instead of 0s
        gr = tf.zeros_like(pnnp)
    gr = cons.backend.real(gr)
    gr = tf.where(tf.math.is_nan(gr), 0.0, gr)
    gmatrix = np.zeros_like(nnp)
    for i, j in enumerate(preset):
        gmatrix[i, j] = gr[i]
    gmatrix = tf.constant(gmatrix)
    return loss, gmatrix


## functions for quantum Hamiltonian QAOA with tensorflow quantum backend
## deprecated tfq related vags

try:
    v = sympy.symbols("v_{0:64}")
    vv = sympy.symbols(["v_" + str(i) + "_0:32" for i in range(32)])

    # symbol pool
    def double_qubits_initial() -> Iterator[Sequence[Any]]:
        while True:
            yield [
                cirq.Circuit(
                    [
                        cirq.rx(0.0)(cirq.GridQubit(0, 0)),
                        cirq.rx(0.0)(cirq.GridQubit(1, 0)),
                    ]
                ),  # 00 +xx +zz
                cirq.Circuit(
                    [cirq.X(cirq.GridQubit(1, 0)), cirq.rx(0.0)(cirq.GridQubit(0, 0))]
                ),  # 01 +xx -zz
                cirq.Circuit(
                    [cirq.X(cirq.GridQubit(0, 0)), cirq.rx(0.0)(cirq.GridQubit(1, 0))]
                ),  # 10 -xx +zz
                cirq.Circuit(
                    [cirq.X(cirq.GridQubit(0, 0)), cirq.X(cirq.GridQubit(1, 0))]
                ),  # 11 -xx -zz
            ]

    def GHZ_vag_tfq(
        gdata: Any,
        nnp: Tensor,
        preset: Sequence[int],
        verbose: bool = False,
        index: Tuple[int, int, int, int, int, int, int, int] = (1, 1, 1, 0, 0, 1, 0, 0),
    ) -> Tuple[Tensor, Tensor]:
        # gdata = quantum_circuit

        circuit = cirq.Circuit()
        cset = get_op_pool()
        for j in preset:
            circuit.append(cset[j])
        input_circuits = [c + circuit for c in gdata]
        measurements = [
            cirq.Z(cirq.GridQubit(0, 0)) * cirq.Z(cirq.GridQubit(1, 0)),
            cirq.X(cirq.GridQubit(0, 0)) * cirq.X(cirq.GridQubit(1, 0)),
        ]
        expl = tfq.layers.Expectation()
        res = expl(input_circuits, operators=measurements)
        if verbose:
            print(res.numpy())
        loss = (
            (-1.0) ** index[0] * res[0, 0]
            + (-1.0) ** index[1] * res[0, 1]
            + (-1.0) ** index[2] * res[1, 0]
            + (-1.0) ** index[3] * res[1, 1]
            + (-1.0) ** index[4] * res[2, 0]
            + (-1.0) ** index[5] * res[2, 1]
            + (-1.0) ** index[6] * res[3, 0]
            + (-1.0) ** index[7] * res[3, 1]
        )
        #     loss = -tf.reduce_sum(tf.abs(res)) # for more general case
        return loss, tf.zeros_like(nnp)

    ## QFT QEM application

    def q(i: int) -> cirq.LineQubit:
        """
        short cut for ``cirq.LineQubit(i)``

        :param i:
        :return:
        """
        return cirq.LineQubit(i)

    @lru_cache()
    def qft_circuit(n: int) -> cirq.Circuit:
        circuit = cirq.Circuit()
        for i in reversed(range(n)):
            circuit.append(cirq.H(q(i)))
            for d, j in enumerate(reversed(range(i))):
                circuit.append(
                    cirq.ControlledGate(cirq.Z ** (1 / 2 ** (d + 1)))(q(j), q(i))
                )

        return circuit

    def gapfilling(circuit: cirq.Circuit, placeholder: Sequence[Any]) -> cirq.Circuit:
        """
        Fill single qubit gates according to placeholder on circuit

        :param circuit:
        :param placeholder:
        :return:
        """
        n_circuit = cirq.Circuit()
        all_qubits = sorted(circuit.all_qubits())
        i = 0
        for m in circuit.moments:
            n_circuit.append(m)
            occupied_qubits = set()
            for g in m:
                for q in g.qubits:
                    occupied_qubits.add(q)
            for q in all_qubits:
                if q not in occupied_qubits:
                    if placeholder[i] != cirq.I:
                        n_circuit.append(placeholder[i](q))
                    i += 1
        return n_circuit

    def noisyfy(
        circuit: cirq.Circuit,
        error_model: str = "bit_flip",
        p_idle: float = 0.2,
        p_sep: float = 0.02,
    ) -> cirq.Circuit:
        noise_circuit = cirq.Circuit()
        error = getattr(cirq, error_model)
        all_qubits = circuit.all_qubits()
        for m in circuit.moments:
            noise_circuit.append(m)
            occupied_qubits = set()
            for g in m:
                for q in g.qubits:
                    occupied_qubits.add(q)
            for q in all_qubits:
                if q not in occupied_qubits:
                    noise_circuit.append(error(p_idle)(q))
            noise_circuit.append(error(p_sep).on_each(*all_qubits))
        return noise_circuit

    def unitary_design_block(circuit: cirq.Circuit, n: int) -> cirq.Circuit:
        """
        random Haar measure approximation

        :param circuit: cirq.Circuit, empty circuit
        :param n: # of qubit
        :return:
        """
        for i in range(n):
            theta = np.random.choice([0, 2 / 3, 4 / 3])
            circuit.append(cirq.ZPowGate(exponent=theta)(q(i)))
        for i in range(n - 1):
            for j in range(i + 1, n):
                if np.random.choice([0, 1]) < 0.5:
                    circuit.append(cirq.CZ(q(i), q(j)))
        for i in range(n):
            circuit.append(cirq.H(q(i)))
        return circuit

    def unitary_design(n: int, l: int = 3) -> cirq.Circuit:
        """
        generate random wavefunction from approximately Haar measure,
        reference:  https://doi.org/10.1063/1.4983266

        :param n: number of qubits
        :param l: repetition of the blocks
        :return:
        """
        circuit = cirq.Circuit()
        for i in range(n):
            circuit.append(
                cirq.H(q(i))
            )  # the first block only final H layer has effect
        for _ in range(l):
            unitary_design_block(circuit, n)
        return circuit

    def qft_qem_vag(
        gdata: Any,
        nnp: Tensor,
        preset: Sequence[int],
        n: int = 3,
        p_idle: float = 0.2,
        p_sep: float = 0.02,
    ) -> Tuple[Tensor, Tensor]:
        # gdata = None
        if gdata is None:
            prepend = unitary_design(n)
        else:
            prepend = gdata
        s = cirq.DensityMatrixSimulator()
        cset = get_op_pool()
        placeholder = [cset[j] for j in preset]
        qftc = qft_circuit(n)

        ideal = prepend + qftc
        pdm = s.simulate(ideal).final_density_matrix
        ncq = gapfilling(qftc, placeholder)
        ncq = noisyfy(ncq, p_idle=p_idle, p_sep=p_sep)
        ncq = prepend + ncq
        ndm = s.simulate(ncq).final_density_matrix
        loss = -cirq.fidelity(pdm, ndm)
        return tf.constant(loss, dtype=tf.float32), tf.zeros_like(nnp)

    @lru_cache()
    def tfim_measurements(
        g: Graph, hzz: float = 1, hx: float = 0, hz: float = 0, one: bool = True
    ) -> Any:
        """
        Hamiltonian for tfim on lattice defined by graph g

        :param g:
        :param hzz:
        :param hx:
        :param hz:
        :param one:
        :return: cirq.PauliSum as operators for tfq expectation layer
        """
        # one=True and False seems give no speed difference
        measurements = []
        qubits = generate_qubits(g)
        for e in g.edges:
            measurements.append(
                hzz
                * g[e[0]][e[1]]["weight"]
                * cirq.Z(qubits[e[0]])
                * cirq.Z(qubits[e[1]])
            )

        if hx != 0:
            for i in range(len(g.nodes)):
                measurements.append(hx * cirq.X(qubits[i]))
        if hz != 0:
            for i in range(len(g.nodes)):
                measurements.append(hz * cirq.Z(qubits[i]))
        if one:
            measurements = reduce(operator.add, measurements)
        return measurements

    @lru_cache()
    def heisenberg_measurements(
        g: Graph, hxx: float = 1.0, hyy: float = 1.0, hzz: float = 1.0, one: bool = True
    ) -> Any:
        """
        Hamiltonian measurements for Heisenberg model on graph lattice g

        :param g:
        :param hxx:
        :param hyy:
        :param hzz:
        :param one:
        :return:
        """
        measurements = []
        qubits = generate_qubits(g)
        for e in g.edges:
            measurements.append(
                hzz
                * g[e[0]][e[1]]["weight"]
                * cirq.Z(qubits[e[0]])
                * cirq.Z(qubits[e[1]])
            )
            measurements.append(
                hxx
                * g[e[0]][e[1]]["weight"]
                * cirq.X(qubits[e[0]])
                * cirq.X(qubits[e[1]])
            )
            measurements.append(
                hyy
                * g[e[0]][e[1]]["weight"]
                * cirq.Y(qubits[e[0]])
                * cirq.Y(qubits[e[1]])
            )
        if one:
            measurements = reduce(operator.add, measurements)
        return measurements

    def quantum_qaoa_vag(
        gdata: Graph,
        nnp: Tensor,
        preset: Sequence[int],
        measurements_func: Optional[Callable[..., Any]] = None,
        **kws: Any,
    ) -> Tuple[Tensor, Tensor]:
        """
        tensorflow quantum backend compare to qaoa_vag which is tensorcircuit backend

        :param gdata:
        :param nnp:
        :param preset:
        :param measurements_func:
        :param kws: kw arguments for measurements_func
        :return:
        """
        if measurements_func is None:
            measurements_func = tfim_measurements
        ep = tfq.layers.Expectation()
        nnp = nnp.numpy()  # real
        pnnp = [nnp[i, j] for i, j in enumerate(preset)]
        pnnp = array_to_tensor(np.array(pnnp))  # complex
        ci = cirq.Circuit()
        cset = get_op_pool()
        for i, j in enumerate(preset):
            if callable(cset[j]):
                cset[j](ci, gdata, v[i])
            else:  # op is a tuple with graph info as (op, g)
                cset[j][0](ci, cset[j][1], v[i])

        with tf.GradientTape() as t:
            t.watch(pnnp)
            loss = ep(
                inputs=[ci],
                symbol_names=v[: len(preset)],
                symbol_values=[pnnp],
                operators=measurements_func(gdata, **kws),
            )[0]
            gr = t.gradient(loss, pnnp)
        if gr is None:
            # if all gates in preset are not trainable, then gr returns None instead of 0s
            gr = tf.zeros_like(pnnp)
        gr = cons.backend.real(gr)
        gr = tf.where(tf.math.is_nan(gr), 0.0, gr)
        gmatrix = np.zeros_like(nnp)
        for i, j in enumerate(preset):
            gmatrix[i, j] = gr[i]
        gmatrix = tf.constant(gmatrix)
        return loss[0], gmatrix

    def quantum_mp_qaoa_vag(
        gdata: Graph,
        nnp: Tensor,
        preset: Sequence[int],
        measurements_func: Optional[Callable[..., Any]] = None,
        **kws: Any,
    ) -> Tuple[Tensor, Tensor]:
        """
        multi parameter for one layer

        :param gdata:
        :param nnp:
        :param preset:
        :param measurements_func:
        :param kws: kw arguments for measurements_func
        :return: loss function, gradient of nnp
        """
        if measurements_func is None:
            measurements_func = tfim_measurements
        assert len(nnp.shape) == 3  # p * c * l
        nnp = nnp.numpy()  # real
        # p, c, l = nnp.shape[0], nnp.shape[1], nnp.shape[2]
        p, l = nnp.shape[0], nnp.shape[2]
        pnnp = np.empty(dtype=np.float32, shape=[p, l])
        for i, j in enumerate(preset):
            pnnp[i, :] = nnp[i, j, :]
        pnnp = array_to_tensor(np.array(pnnp))  # complex
        ci = cirq.Circuit()
        cset = get_op_pool()
        for i, j in enumerate(preset):
            if callable(cset[j]):
                cset[j](ci, gdata, vv[i][:l])
            else:  # op is a tuple with graph info as (op, g)
                cset[j][0](ci, cset[j][1], vv[i][:l])
        ep = tfq.layers.Expectation()
        symbol_names = []
        for i in range(len(preset)):
            symbol_names.extend(vv[i][:l])
        with tf.GradientTape() as t:
            t.watch(pnnp)
            loss = ep(
                inputs=[ci],
                symbol_names=symbol_names,
                symbol_values=[tf.reshape(pnnp, [-1])],
                operators=measurements_func(gdata, **kws),
            )[0]
        gr = t.gradient(loss, pnnp)
        if gr is None:
            # if all gates in preset are not trainable, then gr returns None instead of 0s
            gr = tf.zeros_like(pnnp)
        gr = cons.backend.real(gr)
        gr = tf.where(tf.math.is_nan(gr), 0.0, gr)
        gmatrix = np.zeros_like(nnp)
        for i, j in enumerate(preset):
            gmatrix[i, j, :] = gr[i, :]
        gmatrix = tf.constant(gmatrix)
        return loss[0], gmatrix

except NameError as e:
    logger.warning(e)
    logger.warning("tfq related vags disabled due to missing packages")

## some quantum quantities using tf ops
## deprecated, see new functions implemented in backend agnostic way
## in ``tc.quantum`` module


def entropy(rho: Tensor, eps: float = 1e-12) -> Tensor:
    """
    deprecated, current version in tc.quantum
    """
    lbd = tf.math.real(tf.linalg.eigvals(rho))
    entropy = -tf.math.reduce_sum(lbd * tf.math.log(lbd + eps))
    return tf.math.real(entropy)
    # e = np.linalg.eigvals(rho)
    # eps = 1e-12
    # return -np.real(np.dot(e, np.log(e + eps)))
    # return np.trace(rho@LA.logm(rho))


def renyi_entropy(rho: Tensor, k: int = 2, eps: float = 1e-12) -> Tensor:
    # no matrix power in tf?
    rhok = rho
    for _ in range(k - 1):
        rhok = rhok @ rho
    return 1 / (1 - k) * tf.math.real(tf.linalg.trace(rhok))


def reduced_density_matrix(
    state: Tensor, freedom: int, cut: Union[int, List[int]], p: Optional[Tensor] = None
) -> Tensor:
    """
    deprecated, current version in tc.quantum
    """
    if isinstance(cut, list) or isinstance(cut, tuple):
        traceout = cut
    else:
        traceout = [i for i in range(cut)]
    w = state / tf.linalg.norm(state)
    perm = [i for i in range(freedom) if i not in traceout]
    perm = perm + traceout
    w = tf.reshape(w, [2 for _ in range(freedom)])
    w = tf.transpose(w, perm=perm)
    w = tf.reshape(w, [-1, 2 ** len(traceout)])
    if p is None:
        rho = w @ tf.transpose(w, conjugate=True)
    else:
        rho = w @ tf.linalg.diag(p) @ tf.transpose(w, conjugate=True)
        rho /= tf.linalg.trace(rho)
    return rho


def entanglement_entropy(state: Tensor) -> Tensor:
    """
    deprecated as non tf and non flexible, use the combination of ``reduced_density_matrix`` and ``entropy`` instead.
    """
    state = state.reshape([-1])
    state = state / np.linalg.norm(state)
    t = state.shape[0]
    ht = int(np.sqrt(t))
    square = state.reshape([ht, ht])
    rho = square @ np.conj(np.transpose(square))
    return entropy(rho)


def free_energy(rho: Tensor, h: Tensor, beta: float = 1, eps: float = 1e-12) -> Tensor:
    energy = tf.math.real(tf.linalg.trace(rho @ h))
    s = entropy(rho, eps)
    return tf.math.real(energy - s / beta)


def renyi_free_energy(rho: Tensor, h: Tensor, beta: float = 1) -> Tensor:
    energy = tf.math.real(tf.linalg.trace(rho @ h))
    s = -tf.math.real(tf.math.log(tf.linalg.trace(rho @ rho)))
    return tf.math.real(energy - s / beta)


def taylorlnm(x: Tensor, k: int) -> Tensor:
    dtype = x.dtype
    s = x.shape[-1]
    y = 1 / k * (-1) ** (k + 1) * tf.eye(s, dtype=dtype)
    for i in reversed(range(k)):
        y = y @ x
        if i > 0:
            y += 1 / (i) * (-1) ** (i + 1) * tf.eye(s, dtype=dtype)
    return y


def truncated_free_energy(
    rho: Tensor, h: Tensor, beta: float = 1, k: int = 2, eps: float = 1e-12
) -> Tensor:
    dtype = rho.dtype
    s = rho.shape[-1]
    tyexpand = rho @ taylorlnm(rho - tf.eye(s, dtype=dtype), k - 1)
    renyi = -tf.math.real(tf.linalg.trace(tyexpand))
    energy = tf.math.real(tf.linalg.trace(rho @ h))
    print(energy, renyi, renyi / beta)
    return tf.math.real(energy - renyi / beta)


def trace_distance(rho: Tensor, rho0: Tensor, eps: float = 1e-12) -> Tensor:
    d2 = rho - rho0
    d2 = tf.transpose(d2, conjugate=True) @ d2
    lbds = tf.math.real(tf.linalg.eigvals(d2))
    return 0.5 * tf.reduce_sum(tf.sqrt(lbds + eps))


def fidelity(rho: Tensor, rho0: Tensor) -> Tensor:
    rhosqrt = tf.linalg.sqrtm(rho)
    return tf.math.real(tf.linalg.trace(tf.linalg.sqrtm(rhosqrt @ rho0 @ rhosqrt)) ** 2)


def gibbs_state(h: Tensor, beta: float = 1) -> Tensor:
    rho = tf.linalg.expm(-beta * h)
    rho /= tf.linalg.trace(rho)
    return rho


def double_state(h: Tensor, beta: float = 1) -> Tensor:
    rho = tf.linalg.expm(-beta / 2 * h)
    state = tf.reshape(rho, [-1])
    norm = tf.linalg.norm(state)
    return state / norm


def correlation(m: Tensor, rho: Tensor) -> Tensor:
    return tf.math.real(tf.linalg.trace(rho @ m))
